"""Tests for core signal functions."""

import math
import pytest
import itertools
import numpy as np
import dask.array as da
import astropy.units as u
import pulsarbat as pb
from astropy.time import Time


def assert_equal_signals(x, y):
    assert np.allclose(np.array(x), np.array(y), atol=1e-6)
    assert Time.isclose(x.start_time, y.start_time)
    assert u.isclose(x.sample_rate, y.sample_rate)


def assert_equal_radiosignals(x, y):
    assert_equal_signals(x, y)
    assert u.allclose(x.channel_freqs, y.channel_freqs)


class TestConcatenate:
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_basic(self, use_dask):
        """Basic concatenation functionality."""
        if use_dask:
            f = da.random.standard_normal
        else:
            f = np.random.standard_normal

        shape = (16, 16)
        z = pb.Signal(f(shape), sample_rate=1 * u.Hz, start_time=Time.now())

        x, y = z[:10], z[10:]
        for axis in [0, "time"]:
            z2 = pb.concatenate([x, y], axis=axis)
            assert_equal_signals(z, z2)

            with pytest.raises(ValueError):
                _ = pb.concatenate([y, x], axis=axis)

        y.sample_rate = x.sample_rate * 2
        with pytest.raises(ValueError):
            _ = pb.concatenate([x, y], axis=0)

        x, y = z[:, :10], z[:, 10:]
        z2 = pb.concatenate([x, y], axis=1)
        assert_equal_signals(z, z2)

        with pytest.raises(TypeError):
            _ = pb.concatenate([x, y], axis="freq")

        z = pb.RadioSignal(
            f(shape),
            sample_rate=1 * u.Hz,
            start_time=Time.now(),
            chan_bw=1 * u.MHz,
            center_freq=1 * u.GHz,
        )

        x, y = z[:, :10], z[:, 10:]
        for axis in [1, "freq"]:
            z2 = pb.concatenate([x, y], axis=axis)
            assert_equal_radiosignals(z, z2)

            with pytest.raises(ValueError):
                _ = pb.concatenate([y, x], axis=axis)

        y.chan_bw = x.chan_bw * 2
        with pytest.raises(ValueError):
            _ = pb.concatenate([x, y], axis=1)

    def test_valid_signal(self):
        """
             🡒 t
           ┏━━━┳━━━┓
           ┃ A ┃ B ┃
         ⭣ ┣━━━╋━━━┫
         x ┃ C ┃ D ┃
           ┗━━━┻━━━┛
        valid along t = AB, AD, CB, CD
        valid along x = AA, AC, BB, BD, CC, CA, DD, DB
        """
        shape = (16, 16)
        z = pb.Signal(
            np.random.default_rng().standard_normal(shape),
            sample_rate=1 * u.Hz,
            start_time=Time.now(),
        )

        A, B, C, D = z[:8, :8], z[8:, :8], z[:8, 8:], z[8:, 8:]

        y = pb.concatenate(
            [pb.concatenate([A, B], axis=0), pb.concatenate([C, D], axis=0)], axis=1
        )
        assert_equal_signals(z, y)

        y = pb.concatenate(
            [pb.concatenate([A, C], axis=1), pb.concatenate([B, D], axis=1)], axis=0
        )
        assert_equal_signals(z, y)

        for X, Y in itertools.product([A, B, C, D], repeat=2):
            time_pairs = [(A, B), (A, D), (C, B), (C, D)]

            if any(X is M and Y is N for M, N in time_pairs):
                _ = pb.concatenate([X, Y], axis=0)
            else:
                with pytest.raises(ValueError):
                    _ = pb.concatenate([X, Y], axis=0)

            freq_pairs = [
                (A, A),
                (A, C),
                (B, B),
                (B, D),
                (C, C),
                (C, A),
                (D, D),
                (D, B),
            ]

            if any(X is M and Y is N for M, N in freq_pairs):
                _ = pb.concatenate([X, Y], axis=1)
            else:
                with pytest.raises(ValueError):
                    _ = pb.concatenate([X, Y], axis=1)

    def test_valid_radiosignal(self):
        """
             🡒 t
           ┏━━━┳━━━┓
           ┃ A ┃ B ┃
         ⭣ ┣━━━╋━━━┫
         f ┃ C ┃ D ┃
           ┗━━━┻━━━┛
        valid along t = AB, CD
        valid along f = AC, BD
        """
        shape = (16, 16)
        z = pb.RadioSignal(
            np.random.default_rng().standard_normal(shape),
            sample_rate=1 * u.Hz,
            start_time=Time.now(),
            chan_bw=1 * u.MHz,
            center_freq=1 * u.GHz,
        )

        A, B, C, D = z[:8, :8], z[8:, :8], z[:8, 8:], z[8:, 8:]

        y = pb.concatenate(
            [pb.concatenate([A, B], axis="time"), pb.concatenate([C, D], axis="time")],
            axis="freq",
        )
        assert_equal_radiosignals(z, y)

        y = pb.concatenate(
            [pb.concatenate([A, C], axis="freq"), pb.concatenate([B, D], axis="freq")],
            axis="time",
        )
        assert_equal_radiosignals(z, y)

        for X, Y in itertools.product([A, B, C, D], repeat=2):
            time_pairs = [(A, B), (C, D)]

            if any(X is M and Y is N for M, N in time_pairs):
                _ = pb.concatenate([X, Y], axis=0)
            else:
                with pytest.raises(ValueError):
                    _ = pb.concatenate([X, Y], axis=0)

            freq_pairs = [(A, C), (B, D)]

            if any(X is M and Y is N for M, N in freq_pairs):
                _ = pb.concatenate([X, Y], axis=1)
            else:
                with pytest.raises(ValueError):
                    _ = pb.concatenate([X, Y], axis=1)

    def test_start_time_none(self):
        shape = (32, 32)
        k = (0, 13, 19, 19, 23, 32)

        z = pb.Signal(
            np.random.default_rng().standard_normal(shape),
            sample_rate=1 * u.Hz,
            start_time=Time.now(),
        )

        for c in range(len(k) - 1):
            zs = [z[i:j] for i, j in zip(k, k[1:])]

            for j, x in enumerate(zs):
                if j != c:
                    x.start_time = None

            y = pb.concatenate(zs, axis="time")
            assert_equal_signals(z, y)

        zs = [z[i:j] for i, j in zip(k, k[1:])]
        for x in zs:
            x.start_time = None
        y = pb.concatenate(zs, axis="time")
        assert y.start_time is None

    def test_edge_cases(self):
        shape = (16, 16)

        z = pb.Signal(
            np.random.default_rng().standard_normal(shape),
            sample_rate=1 * u.Hz,
            start_time=Time.now(),
        )

        with pytest.raises(ValueError):
            _ = pb.concatenate([])

        with pytest.raises(TypeError):
            _ = pb.concatenate([z.data])

        with pytest.raises(TypeError):
            _ = pb.concatenate([z[:8], z[8:].data])


class TestTimeShift:
    @pytest.mark.parametrize("use_dask", [True, False])
    @pytest.mark.parametrize("use_complex", [True, False])
    def test_int_roll(self, use_dask, use_complex):
        if use_dask:
            f = da.random.standard_normal
        else:
            f = np.random.standard_normal

        shape = (4096, 4, 2)

        if use_complex:
            x = (f(shape) + 1j * f(shape)).astype(np.complex128)
        else:
            x = f(shape).astype(np.float64)

        z = pb.Signal(x, sample_rate=1 * u.kHz, start_time=Time.now())

        for n in [1, 10, 55, 211]:
            assert_equal_signals(z[:-n], pb.time_shift(z, n))
            assert_equal_signals(z[:-n], pb.time_shift(z, n * u.ms))

            assert_equal_signals(z[n:], pb.time_shift(z, -n))
            assert_equal_signals(z[n:], pb.time_shift(z, -n * u.ms))

    def test_subsample_roll(self):
        def impulse(N, t0):
            """Generate noisy impulse at t0, with given S/N."""
            n = (np.arange(N) - N // 2) / N
            x = np.exp(-2j * np.pi * t0 * n)
            return np.fft.ifft(np.fft.ifftshift(x)).astype(np.complex128)

        N = 1024
        for shift in [16.5, 32.25, 50.1, 60.9, 466.666]:
            imp1 = impulse(N, shift)
            imp2 = impulse(N - math.ceil(shift), 0)
            x = pb.Signal(imp1, sample_rate=1 * u.kHz, start_time=Time.now())
            y = pb.time_shift(x, -shift)
            assert np.allclose(np.array(y), imp2)
            z = pb.time_shift(x, -shift * u.ms)
            assert np.allclose(np.array(z), imp2)


class TestFastLen:
    def test_fast(self):
        for N in [4096, 4100, 4111]:
            x = np.arange(N, dtype=np.float64)
            z = pb.Signal(x, sample_rate=1 * u.Hz)
            y = pb.fast_len(z)
            assert len(y) == 4096
            assert np.allclose(y.data, np.arange(4096))


class TestSignalTransform:
    @pytest.mark.parametrize("arange", [np.arange, da.arange])
    def test_median_filter(self, arange):
        from scipy.ndimage import median_filter

        data = arange(9).reshape(-1, 3)
        res = median_filter(arange(9).reshape(-1, 3), size=3, mode='constant')

        sig_med_filt = pb.signal_transform(median_filter)

        t0 = Time.now()
        z = pb.Signal(data, sample_rate=1*u.Hz, start_time=t0)

        x = type(z).like(z, res)
        y = sig_med_filt(z, size=3, mode='constant')

        assert isinstance(y.data, type(z.data))
        assert_equal_signals(x, y)

        kw = dict(center_freq=1*u.GHz, chan_bw=10*u.MHz)
        y = sig_med_filt(z, signal_type=pb.IntensitySignal, signal_kwargs=kw,
                         size=3, mode='constant')

        assert isinstance(y, pb.IntensitySignal)
        assert y.center_freq == kw['center_freq']
        assert y.chan_bw == kw['chan_bw']

        with pytest.raises(TypeError):
            _ = sig_med_filt(z, signal_type=np.ndarray, size=3, mode='constant')



"""Experimental routines."""

# flake8: noqa

import numpy as np
import astropy.units as u
from astropy.time import Time
import pulsarbat as pb


__all__ = [
    "stft",
    "istft",
]


def stft(z, /, window="boxcar", nperseg=256, noverlap=0, nfft=None):
    """Performs a short-time Fourier transform on a baseband signal.

    Behaves the same as `scipy.signal.stft`. Currently, only supports
    `window='boxcar'`, `noverlap=0` and `nfft = None`. This results in
    the critically-sampled perfect reconstruction STFT - using a
    rectangular window function with no overlap.

    For now, users should only use the `z` and `nperseg` arguments.

    When fully implemented, should behave exactly as `scipy.signal.stft`
    with support for different window and overlap configurations, and
    lazy execution via dask.
    """
    if window != "boxcar" or noverlap != 0 or nfft != None:
        return NotImplemented

    if not isinstance(z, pb.BasebandSignal):
        raise ValueError("z must be a BasebandSignal.")

    if nfft is None:
        nfft = nperseg

    nperseg, noverlap, nfft = int(nperseg), int(noverlap), int(nfft)
    z = z[: len(z) - len(z) % nperseg, :]

    new_shape = (-1, nperseg) + z.sample_shape
    x = z.data.reshape(new_shape)
    x = x.swapaxes(1, 2)

    x = pb.fft.fft(x, axis=2, n=nfft)
    x = np.fft.fftshift(x, axes=(2,))

    out_shape = (x.shape[0], -1) + x.shape[3:]
    x = x.reshape(out_shape)
    x /= nperseg

    falign = "center" if nfft % 2 else "bottom"
    return type(z).like(z, x, sample_rate=z.sample_rate / nfft, freq_align=falign)


def istft(z, /, window="boxcar", nperseg=256, noverlap=0, nfft=None):
    """Performs an inverse short-time Fourier transform.

    Behaves the same as `scipy.signal.istft`. Currently, only supports
    `window='boxcar'`, `noverlap=0` and `nfft = None`.

    For now, users should only use the `z` and `nperseg` arguments.

    When fully implemented, should behave exactly as `scipy.signal.istft`
    with support for different window and overlap configurations, and
    lazy execution via dask.
    """
    if window != "boxcar" or noverlap != 0 or nfft != None:
        return NotImplemented

    if not isinstance(z, pb.BasebandSignal):
        raise ValueError("z must be a BasebandSignal.")

    if nfft is None:
        nfft = nperseg

    nperseg, noverlap, nfft = int(nperseg), int(noverlap), int(nfft)

    new_shape = (len(z), -1, nfft) + z.sample_shape[1:]
    x = z.data.reshape(new_shape)
    x *= nperseg

    x = x.swapaxes(1, 2)
    x = np.fft.ifftshift(x, axes=(1,))
    x = pb.fft.ifft(x, axis=1, n=nfft)
    x = x[:, :nperseg]

    out_shape = (-1,) + x.shape[2:]
    x = x.reshape(out_shape)

    return type(z).like(z, x, sample_rate=z.sample_rate * nfft, freq_align="center")



"""Tests for polarization features."""

import pytest
import numpy as np
import dask.array as da
import astropy.units as u
import pulsarbat as pb


@pytest.mark.parametrize("use_dask", [True, False])
@pytest.mark.parametrize("pol_type", ["linear", "circular"])
def test_pol_reversibility(use_dask, pol_type):
    shape = (4096, 16, 2)
    kw = {"sample_rate": 1 * u.MHz, "center_freq": 1 * u.GHz}

    p = da if use_dask else np
    sig = p.exp(1j * p.random.uniform(-np.pi, +np.pi, shape))

    z = pb.DualPolarizationSignal(sig, pol_type=pol_type, **kw)

    if pol_type == "linear":
        x = z.to_linear()
        y = z.to_circular().to_linear()
    else:
        x = z.to_circular()
        y = z.to_linear().to_circular()

    for a in [x, y]:
        assert isinstance(a.data, type(sig))
        assert a.pol_type == pol_type
        assert np.allclose(np.array(z), np.array(a))


@pytest.mark.parametrize("use_dask", [True, False])
def test_stokes(use_dask):
    kw = {"sample_rate": 1 * u.MHz, "center_freq": 1 * u.GHz}

    x = np.array(
        [[[1 + 1j, 2 + 1j]], [[3 + 0j, 0 + 4j]], [[0 + 2j, 3 + 1j]]],
        dtype=np.complex128,
    )

    if use_dask:
        x = da.from_array(x)

    lin_stokes = np.array([[[7, -3, 6, -2]], [[25, -7, 0, 24]], [[14, -6, 4, -12]]])

    cir_stokes = np.array([[[7, 6, -2, -3]], [[25, 0, 24, -7]], [[14, 4, -12, -6]]])

    for pol_type, stokes in zip(["linear", "circular"], [lin_stokes, cir_stokes]):

        z = pb.DualPolarizationSignal(x, pol_type=pol_type, **kw)
        y_lin = z.to_linear().to_stokes()
        y_cir = z.to_circular().to_stokes()

        print(pol_type)
        print(y_lin.data)
        print(y_cir.data)

        for y in [y_lin, y_cir]:
            assert isinstance(y, pb.FullStokesSignal)
            assert isinstance(y.data, type(x))
            assert np.allclose(np.asarray(y), stokes)

            for i, s in enumerate(["I", "Q", "U", "V"]):
                for a in [y[s], getattr(y, "stokes" + s)]:
                    assert isinstance(a, pb.IntensitySignal)
                    assert np.allclose(np.asarray(a), stokes[..., i])


def test_stokes_getitem():
    kw = {"sample_rate": 1 * u.MHz, "center_freq": 1 * u.GHz, "chan_bw": 1 * u.MHz}

    x = np.array([[[7, -3, 6, -2]], [[25, -7, 0, 24]], [[14, -6, 4, -12]]])

    z = pb.FullStokesSignal(x, **kw)

    y = z[:2]
    assert isinstance(y, type(z))
    assert np.allclose(np.array(y), x[:2])

    with pytest.raises(KeyError):
        _ = z["fish"]



"""The setup script."""

from setuptools import setup

setup(
    name="pulsarbat",
    install_requires=[
        "astropy >= 5.1",
        "numpy >= 1.22",
        "scipy >= 1.9",
        "baseband >= 4.1.1",
        "dask[array] >= 2022.8.1",
    ],
    tests_require=["pytest >= 7.1.2", "cloudpickle >= 2.1.0"],
)



"""Reader module for pulsarbat."""

# flake8: noqa

from . import _base
from ._base import *

from . import _baseband_readers
from ._baseband_readers import *

__all__ = _base.__all__.copy()
__all__ += _baseband_readers.__all__



"""Dedispersion routines."""

import math
import numpy as np
import astropy.units as u
import pulsarbat as pb
import dask
import dask.array as da


__all__ = [
    "DispersionMeasure",
    "DM",
    "coherent_dedispersion",
    "incoherent_dedispersion",
]


def _transfer_function(coeff, N, dt, center_freq, ref_freq):
    f = center_freq.to(u.Hz) + np.fft.fftfreq(N, dt).to(u.Hz)
    phase = coeff * f * u.cycle * (1 / ref_freq - 1 / f) ** 2
    tf = np.exp(-1j * phase.to_value(u.rad))
    return tf.astype(np.complex64)


class DispersionMeasure(u.SpecificTypeQuantity):
    """Dispersion Measure class (with default units of pc / cm^3)."""

    _equivalent_unit = _default_unit = u.pc / u.cm ** 3
    dispersion_constant = u.s * u.MHz ** 2 * u.cm ** 3 / u.pc / 2.41e-4

    def time_delay(self, f, ref_freq):
        """Time delay of frequencies relative to reference frequency."""
        coeff = self.dispersion_constant * self
        delay = coeff * (1 / f ** 2 - 1 / ref_freq ** 2)
        return delay.to(u.s)

    def sample_delay(self, f, ref_freq, sample_rate):
        """Sample delay of frequencies relative to reference frequency."""
        samples = self.time_delay(f, ref_freq) * sample_rate
        samples = samples.to_value(u.one)
        return samples

    def chirp_function(self, N, dt, center_freq, ref_freq, use_dask=False):
        """Chirp function for coherent dedispersion."""
        coeff = self.dispersion_constant * self
        tf_args = (coeff, N, dt, center_freq, ref_freq)

        if use_dask:
            delayed_tf = dask.delayed(_transfer_function, pure=True)
            chirp = da.from_delayed(
                delayed_tf(*tf_args), dtype=np.complex64, shape=(N,)
            )
        else:
            chirp = _transfer_function(*tf_args)

        return chirp

    def chirp_from_signal(self, z, /, *, ref_freq=None):
        """Returns chirp function to dedisperse given baseband signal."""
        if not isinstance(z, pb.BasebandSignal):
            raise TypeError("Signal must be a BasebandSignal object.")

        ix = tuple(slice(None) if i < 2 else None for i in range(z.ndim))
        N, dt = len(z), z.dt

        if ref_freq is None:
            ref_freq = z.center_freq

        chirps = [
            self.chirp_function(N, dt, f, ref_freq, isinstance(z.data, da.Array))
            for f in z.channel_freqs
        ]

        return np.stack(chirps, axis=1)[ix]


DM = DispersionMeasure


def coherent_dedispersion(z, DM, /, *, ref_freq=None, chirp=None):
    """Coherently dedisperses a baseband signal.

    The given signal will be coherently dedispersed by a given dispersion
    measure (DM). If a reference frequency (``ref_freq``) is not given, the
    center frequency of the signal will be used as reference.

    Optionally, a pre-computed chirp function (``chirp``) can be provided
    as an array. If a chirp is provided, it will not be checked against
    the given DM and reference frequency for correctness.

    The output signal will be cropped on both ends to avoid wrap-around
    artifacts caused by dedispersion. This depends on where the
    reference frequency (``ref_freq``) is compared to the band of the
    signal.

    Parameters
    ----------
    z : BasebandSignal
        The signal to be transformed.
    DM : DispersionMeasure
        Dispersion measure by which to dedisperse ``z``.
    ref_freq : Quantity, optional
        Reference frequency for dedispersion. If None (default), uses
        the center frequency from signal.
    chirp : array-like, optional
        A pre-computed chirp function. Must be a 2-D array with shape
        ``z.shape[:2]``.

    Returns
    -------
    out : BasebandSignal
        The dedispersed signal.
    """
    if not isinstance(z, pb.BasebandSignal):
        raise TypeError("Signal must be a BasebandSignal object.")

    if ref_freq is None:
        ref_freq = z.center_freq

    if chirp is None:
        chirp = DM.chirp_from_signal(z, ref_freq=ref_freq)

    x = pb.fft.ifft(pb.fft.fft(z.data, axis=0) * chirp, axis=0)

    delay_top = DM.sample_delay(z.max_freq, ref_freq, z.sample_rate)
    delay_bot = DM.sample_delay(z.min_freq, ref_freq, z.sample_rate)

    start = math.ceil(-min(0, delay_top, delay_bot))
    stop = x.shape[0] - math.ceil(+max(0, delay_top, delay_bot))

    return type(z).like(z, x)[start:stop]


def incoherent_dedispersion(z, DM, /, *, ref_freq=None):
    """Incoherently dedisperses a signal by a given dispersion measure.

    The output signal will be cropped on both ends to avoid wrap-around
    artifacts caused by dedispersion. This depends on where the
    reference frequency (``ref_freq``) compared to the band of the signal.

    Parameters
    ----------
    z : RadioSignal
        The signal to be transformed.
    DM : DispersionMeasure
        Dispersion measure by which to dedisperse ``z``.
    ref_freq : Quantity, optional
        Reference frequency for dedispersion. If None (default), uses
        the center frequency from signal.

    Returns
    -------
    out : RadioSignal
        The dedispersed signal.
    """
    if not isinstance(z, pb.RadioSignal):
        raise TypeError("Signal must be a RadioSignal object.")

    if ref_freq is None:
        ref_freq = z.center_freq

    delays = DM.sample_delay(z.channel_freqs, ref_freq, z.sample_rate)
    delays = delays.round().astype(np.int64)

    crop_before = -min(0, delays[0], delays[-1])
    delays += crop_before
    N = len(z) - max(delays)

    x = np.stack([z.data[j : j + N, i] for i, j in enumerate(delays)], axis=1)

    new_start = z.start_time
    if crop_before and z.start_time is not None:
        new_start += crop_before * z.dt

    return type(z).like(z, x, start_time=new_start)
